module Day10
  ( Instruction (..),
    parseProgram,
    signalStrength,
    executeProgram,
    draw,
    isSpriteDrawn,
    day10
  )
where

import Control.Arrow ((>>>))
import Data.List.Split (chunksOf)

data Instruction = Addx Int | Noop
  deriving (Show, Eq)

newtype Pixel = Pixel Bool deriving (Eq)

instance Show Pixel where
  show (Pixel True) = "▓"
  show (Pixel False) = "░"


type CRT = String

parseProgram :: String -> [Instruction]
parseProgram = lines >>> map parseProgram'
  where
    parseProgram' :: String -> Instruction
    parseProgram' "noop" = Noop
    parseProgram' line = splitAt 4 >>> snd >>> read >>> Addx $ line


isSpriteDrawn :: Int -> Int -> Bool
isSpriteDrawn x p = abs (position - x) <= 1
  where
    position = p `mod` 40

draw :: [Int] -> CRT
draw registerValues = unlines lastPicture
  where
    pictures = zipWith isSpriteDrawn registerValues [0..]
    lastPicture = concatMap (show . Pixel) >>> chunksOf 40 $ pictures



signalStrength :: [Int] -> Int
signalStrength e = start + sum computedRest
  where
    start = e !! 20 * 20
    rest = drop 20 e
    withCycle = zip [60, 100..] (every 40 rest)
    computedRest = [c*r | (c,r) <- withCycle]

every :: Int -> [a] -> [a]
every n xs = case drop (n-1) xs of
              y : ys -> y : every n ys
              [] -> []

executeProgram :: [Instruction] -> [Int]
executeProgram = scanl (+) 1 . concatMap execOne
  where
    execOne instruction = case instruction of
        Noop   -> [0]
        Addx x -> [0, x]

day10 :: IO ()
day10 = do
  input <- readFile "ressources/day10-input"
  putStrLn "Day10"
  let states = parseProgram >>> executeProgram $ input
  let signal = signalStrength states
  let crt = draw states
  putStrLn ("Signal strength: " ++ show signal)
  putStrLn ("CRT picture: \n" ++ crt)
