{-# LANGUAGE OverloadedStrings #-}

module Day11
  ( runMonkeyMachine,
    mostActiveMonkeys,
    monkeyBusiness,
    day11,
    runPart1Machine,
    runPart2Machine,
    worryLimit,
    divideBy3,
  )
where

import Control.Monad.State.Strict (execState, get, modify', put)
import qualified Control.Monad.State.Strict as St
import Data.Foldable (forM_, toList)
import Data.List (sortOn)
import Data.Ord (Down (..))
import Data.Sequence ((|>))
import qualified Data.Sequence as S
import Data.Text (pack)
import Day11.Parser

day11 :: IO ()
day11 = do
  input <- readFile "ressources/day11-input"
  putStrLn "Day11"
  let monkeys = (monkeyBusiness . mostActiveMonkeys . runPart1Machine . parseMonkeys) (pack input)
  putStrLn ("Part 1 monkey business: " ++ show monkeys)
  let monkeys' = (monkeyBusiness . mostActiveMonkeys . runPart2Machine . parseMonkeys) (pack input)
  putStrLn ("Part 2 monkey business: " ++ show monkeys')

type MonkeyState = St.State (S.Seq Monkey)

runPart1Machine :: [Monkey] -> [Monkey]
runPart1Machine = runMonkeyMachine divideBy3 20

runPart2Machine :: [Monkey] -> [Monkey]
runPart2Machine monkeys = runMonkeyMachine (worryLimit monkeys) 10000 monkeys

worryLimit :: [Monkey] -> WorryReducer
worryLimit monkeys = (`mod` limit)
  where
    limit = product $ map _test monkeys

    
runMonkeyMachine :: WorryReducer -> Int -> [Monkey] -> [Monkey]
runMonkeyMachine eval i monkeys = toList (infiniteMachine !! i)
  where
    infiniteMachine = iterate go $ S.fromList monkeys
    go :: S.Seq Monkey -> S.Seq Monkey
    go m = execState go' m
    go' :: MonkeyState ()
    go' = mapM_ (doMonkey eval) [0 .. length monkeys - 1]

mostActiveMonkeys :: [Monkey] -> [Monkey]
mostActiveMonkeys = take 2 . sortOn (Down . _inspectionCount)

monkeyBusiness :: [Monkey] -> Int
monkeyBusiness = product . map _inspectionCount

-- update a sequence of monkeys according to the moves of the monkey at position idx
doMonkey :: WorryReducer -> Int -> MonkeyState ()
doMonkey reducer idx = do
  horde <- get
  let m = horde `S.index` idx
  let targets = fmap (evalMonkey reducer m) (_items m)
  let m' = m {_items = S.empty, _inspectionCount = _inspectionCount m + S.length (_items m)}
  put $ S.update idx m' horde
  forM_ targets $ \movement ->
    modify' (move movement)

-- update other monkeys according to a list of calculated moves
move :: (Int, Int) -> S.Seq Monkey -> S.Seq Monkey
move (to, item) horde
  | item < 0 = error ("Item worry level overflow: " ++ show (to, item))
  | otherwise =
    let m = horde `S.index` to
        m' = m {_items = _items m |> item}
     in S.update to m' horde

type WorryReducer = Int -> Int

divideBy3 :: WorryReducer
divideBy3 = flip div 3

evalMonkey :: WorryReducer -> Monkey -> Int -> (Int, Int)
evalMonkey reducer monkey item = (target, newLevel)
  where
    afterOp = case _operation monkey of
      Add Old -> item * 2
      Add (Fixed x) -> item + x
      Multiply Old -> item * item
      Multiply (Fixed x) -> item * x
    newLevel = reducer afterOp
    isDivisible = newLevel `mod` _test monkey == 0
    target =
      if isDivisible
        then _ifDivisible monkey
        else _otherwise monkey
