{-# LANGUAGE OverloadedStrings #-}
module Day11.Parser (
  parseMonkeys,
  Monkey (..),
  Operation (..),
  Op (..),
)
 where

import Text.Megaparsec
import Text.Megaparsec.Char
import Data.Text (Text)
import Data.Void (Void)
import Control.Monad (void)
import qualified Data.Sequence as S

type Parser = Parsec Void Text

data Op = Fixed Int | Old
  deriving (Show, Eq)

data Operation = Add Op | Multiply Op
  deriving (Show, Eq)

data Monkey = Monkey
  { _nr :: Int
  , _items :: S.Seq Int
  , _operation :: Operation
  , _test :: Int
  , _ifDivisible :: Int
  , _otherwise :: Int
  , _inspectionCount :: Int }
  deriving (Show, Eq)

parseMonkeys :: Text -> [Monkey]
parseMonkeys input = case parse (monkey `sepBy` void eol) "" input of
  Left err -> error (show err)
  Right result -> result

monkey :: Parser Monkey
monkey = do
  void (string "Monkey ")
  nr <- many numberChar
  void (char ':')
  void eol
  startingItems <- items
  operation <- op
  d <- divisor
  ifDiv <- ifDivisible
  ifOther <- ifOtherwise
  return Monkey
   { _nr=read nr
   , _items=S.fromList startingItems
   , _operation=operation
   , _test=d
   , _ifDivisible=ifDiv
   , _otherwise=ifOther
   , _inspectionCount=0 }

ifDivisible :: Parser Int
ifDivisible = do
  x <- read <$> (string "    If true: throw to monkey " >> many numberChar)
  void eol
  return x

ifOtherwise :: Parser Int
ifOtherwise = do
  x <- read <$> (string "    If false: throw to monkey " >> many numberChar)
  void eol
  return x

divisor :: Parser Int
divisor = do
  d <- read <$> (string "  Test: divisible by " >> many numberChar)
  void eol
  return d

items :: Parser [Int]
items = do
  void (string "  Starting items: ")
  startingItems <- sepBy1 (many numberChar) (string ", ")
  void eol
  return (map read startingItems)

op :: Parser Operation
op = do
  o <- void (string "  Operation: new = old ") >> addOp <|> mulOp
  void eol
  return o
  where
    old = string "old" >> return Old
    num = Fixed . read <$> many numberChar
    addOp = do
      void (string "+ ")
      Add <$> (old <|> num)
    mulOp = do
      void (string "* ")
      Multiply <$> (old <|> num)
