module Day5.Shared (
  StackIndex (..),
  Operation (..),
  Stack (..),
  Row,
  Content,
  asInt,
  content,
  parseOperation,
  parseStacks,
  parseRow,
  combineRows,
  parseInput,
  zipWithPadding,
  findTopOfStacks
) where

import qualified Data.Sequence as S
import Data.Sequence (Seq (..))

newtype StackIndex = StackIndex Int deriving (Show, Eq, Ord)
asInt :: StackIndex -> Int
asInt (StackIndex i) = i - 1

data Operation = Operation
                    { count :: Int
                    , from :: StackIndex
                    , to :: StackIndex} deriving (Show, Eq)

type Content = Char
newtype Stack = Stack [Content] deriving (Show, Eq)

content :: Stack -> [Content]
content (Stack c) = c

instance Semigroup Stack where
  (Stack x) <> (Stack y) = Stack (x ++ y)

instance Monoid Stack where
  mempty = Stack []

type Row = S.Seq Stack

parseOperation :: String -> Operation
parseOperation line = Operation {
 count = read $ parts !! 1,
  from = StackIndex (read $ parts !! 3),
  to = StackIndex (read $ parts !! 5)
  }
  where
    parts = words line

parseStacks :: [String] -> Row
parseStacks input = foldl combineRows S.empty rows
  where
    rows = map parseRow input

parseRow :: String -> Row
parseRow ('[':c:']':' ':cs) = Stack [c] :<| parseRow cs
parseRow (' ':' ':' ':' ':cs) = Stack [] :<| parseRow cs
parseRow ['[', c, ']'] = S.singleton $ Stack [c]
parseRow [' ', ' ', ' '] = S.singleton $ Stack []
parseRow (' ':'1':_) = S.empty -- skip last line
parseRow cs = error ("parseRow: unexpected Row \"" ++ cs ++ "\"")

combineRows :: Row -> Row -> Row
combineRows S.Empty row = row
combineRows row S.Empty = row
combineRows up down = S.fromList [x <> y | (x,y) <- zipWithPadding up down]

zipWithPadding :: (Monoid a, Monoid b) => S.Seq a -> S.Seq b -> [(a,b)]
zipWithPadding (x :<| xs)  (y :<| ys)  = (x,y) : zipWithPadding xs ys
zipWithPadding S.Empty (y :<| ys)  = (mempty, y) : zipWithPadding S.empty ys
zipWithPadding (x :<| xs)  S.Empty = (x, mempty) : zipWithPadding xs S.empty
zipWithPadding S.Empty  S.Empty = []

parseInput :: String -> (Row, [Operation])
parseInput input = (parseStacks stackInput, map parseOperation (drop 1 operationInput))
  where
    (stackInput, operationInput) = break ("" ==) $ lines input


findTopOfStacks :: Row -> Seq Content
findTopOfStacks row = head . content <$> row
